#include <iostream.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <errno.h>
#include <signal.h>

#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/soundcard.h>

#include <qpalette.h>
#include <qcolor.h>
#include <qlayout.h>
#include <qlabel.h>

#include "qntv.h"
#include "videodev.h"

#include <X11/Xlib.h>

// Capture size (scales to fit window - don't mess with this)
#define MY_WIDTH 320
#define MY_HEIGHT 240

typedef struct tag_vimage
{
	struct video_buffer     vidbuf;
	char                    *data;
} VIMAGE;

// global variables
int vol = 75;			// current sound volume
int cur_width = MY_WIDTH;	// current video width
int cur_height = MY_HEIGHT;	// current video height
int mixer_fd = -1;		// the mixer file descriptor
int vid_fd = -1;		// the video input file descriptor
int source;			// mixer input source
unsigned supported;		// soundcard supported features bitfield

Qntv::Qntv()
{
	char buf[8];
	
	// the video window, filled with our chroma color (0xff00ff)
	QLabel *box = new QLabel(this,"box",0U);
	box->setPalette(QPalette(QColor(255,0,255)));
	box->resize(MY_WIDTH,MY_HEIGHT);
	
	// a vertical layout to hold our window and buttons
	QVBoxLayout *vbox = new QVBoxLayout(this);
	vbox->addWidget(box,0,0);
	
	// the slider
	QHBoxLayout *hbox = new QHBoxLayout();
	vbox->addLayout(hbox,0);
	QLabel *lb = new QLabel("Volume: ",this,0);
	lb->setMinimumSize(lb->sizeHint());
	lb->setMaximumHeight(15);
	volSlider = new QSlider(0,100,5,75,QSlider::Horizontal,this,"vslider");
	volSlider->setMaximumHeight(15);
	hbox->addWidget(lb,0,0);
	hbox->addWidget(volSlider,1,0);
	
	connect(volSlider, SIGNAL(valueChanged(int)), SLOT(volChangeCallback(int)));
}

Qntv::~Qntv()
{
}

void Qntv::MoveVidcapWindow(int x, int y)
{
	struct video_window vw = 
		{x,y,cur_width,cur_height,0xff00ff,0,0};
	ioctl(vid_fd, VIDIOC_S_WIN, &vw);
}

void Qntv::setSource(int id)
{
	source = 1 << id;
	if (ioctl (mixer_fd, SOUND_MIXER_WRITE_RECSRC, &source) < 0)
		perror ("Mixer.set_source()");
}

void Qntv::setVolume(int mixer_id, int volume)
{
	if (ioctl (mixer_fd, MIXER_WRITE (mixer_id), &volume) < 0)
		perror ("Mixer.set_mixer()");
}

void Qntv::volChangeCallback(int v)
{
	vol = v;
	setVolume(SOUND_MIXER_IMIX, (vol<<8)+vol);
}

// Set mixer input source to 0, zero rec mon volume and close descriptors
void shutdown(int sig)
{
	int parm = 0;
	ioctl(mixer_fd, MIXER_WRITE (SOUND_MIXER_IMIX), &parm);
	ioctl(mixer_fd, SOUND_MIXER_WRITE_RECSRC, &parm);
	close(mixer_fd);
	close(vid_fd);
	exit(0);
}

// Clean up and exit
void Qntv::closeEvent(QCloseEvent *e)
{
	shutdown(0);
}

// Handle window movement
void Qntv::moveEvent(QMoveEvent *e)
{
	QPoint pos = e->pos();
	this->MoveVidcapWindow(pos.x(), pos.y());
	this->resize(cur_width, cur_height);
	this->repaint();
}

// Handle window resizing
void Qntv::resizeEvent(QResizeEvent *e)
{
	QPoint pos = this->pos();
	QSize sz = e->size();
	cur_width = sz.width();
	cur_height = sz.height();
	this->MoveVidcapWindow(pos.x(),pos.y());
	this->repaint();
}

int main(int argc, char **argv)
{
	int err,i;
	struct video_format fmt;
	struct video_parm parm;
	struct video_standard std;
	VIMAGE vimage;
	struct video_framebuffer vf;
	Display *display;

	// Die gracefully on ctrl-c and term signals
	signal(SIGINT, &shutdown);
	signal(SIGTERM, &shutdown);

	// Open the default X display
	display = XOpenDisplay(NULL);
	if(!display) {
		printf("Can't open X display\n\n");
		return -1;
	}

	// Set up the Qt windows and app context
	QApplication *app = new QApplication(argc,argv);
	Qntv *tv = new Qntv();
	app->setMainWidget(tv);
	tv->show();
	
	// Open video device
	vid_fd = open("/dev/video0", O_RDONLY);
	if(vid_fd < 0) {
		perror("Vidcap.open");
		return -1;
	}
	
	// Open mixer and set volume
	mixer_fd = open("/dev/mixer", O_RDWR);
	if(mixer_fd < 0) {
		perror("Mixer.open");
		return -1;
	}

	if(ioctl(mixer_fd, SOUND_MIXER_READ_DEVMASK, &supported) < 0)
		supported = 0xffff;
	tv->setSource(6); // turn on record monitor on line in
	tv->setVolume(SOUND_MIXER_IMIX, (vol<<8)+vol); // set source volume

//->    Set video params 0=MXC, 1=COMPOSITE, 2=SVIDEO
	ioctl(vid_fd, VIDIOC_G_PARM, &parm);
	parm.input = 0;
	err = ioctl(vid_fd, VIDIOC_S_PARM, &parm);
	
//->    Set video standard VIDEO_STD_PAL or VIDEO_STD_NTSC
	err = ioctl(vid_fd, VIDIOC_S_STD, &std);

//->    Image width, height, pixel format
	ioctl(vid_fd, VIDIOC_G_FMT, &fmt);
        fmt.width = MY_WIDTH;
        fmt.height = MY_HEIGHT;
        fmt.pixelformat = PIX_FMT_RGB565;
        err = ioctl(vid_fd, VIDIOC_S_FMT, &fmt);
        if (err) {
		perror("Vidcap:set_params");
                return -1;
	}

	ioctl(vid_fd, VIDIOC_G_FBUF, &vf);
	vf.flags = FBUF_FLAG_CHROMAKEY | FBUF_FLAG_OVERLAY;
	ioctl(vid_fd, VIDIOC_S_FBUF, &vf);
	ioctl(vid_fd, VIDIOC_PREVIEW, 1);

	vimage.data = (char *)malloc(fmt.sizeimage);
	if(vimage.data == 0) {
		cout << "Unable to allocate memory for video buffer." << endl;
		return -1;
	}

	return app->exec();
}

#include "qntv.moc"
